/*
 * Main program entry point - read the command line options, then perform
 * the appropriate actions.
 *
 * Copyright 2004 Andrew Wood, distributed under the Artistic License.
 */

#include "config.h"
#include "options.h"
#include "message.h"
#include "spam.h"
#include "database.h"
#include "log.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef HAVE_MCHECK_H
#include <mcheck.h>
#endif


/*
 * Process command-line arguments and set option flags, then call functions
 * to initialise, and finally enter the main loop.
 */
int main(int argc, char **argv)
{
	char buffer[1024];		 /* RATS: ignore (checked all) */
	char filename[1024];		 /* RATS: ignore (checked all) */
	opts_t options;
	int retcode = 0;
	int needwrite = 0;
	int got, sent, totsent;
	double score;
	qdb_t dbr1, dbr2, dbr3, dbw;
	char *home;
	msg_t msg = NULL;
	int fd = -1;

#ifdef HAVE_MCHECK_H
	if (getenv("MALLOC_TRACE"))	    /* RATS: ignore (value unused) */
		mtrace();
#endif

#ifdef ENABLE_NLS
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
#endif

	options = parse_options(argc, argv);
	if (!options)
		return 1;

	if (options->do_nothing) {
		options->destructor(options);
		return 0;
	}

	/*
	 * If we're not dumping, restoring, token listing, pruning,
	 * training, benchmarking, merging, or allow-list updating/querying
	 * (given an email address), we need to read a message from standard
	 * input before proceeding.
	 */
	if (!options->dump && !options->restore
	    && !options->train && !options->benchmark
	    && !options->tokens && !options->prune
	    && !options->merge && ((!options->emailonly)
				   || (options->emailonly
				       &&
				       (strcmp(options->emailonly, "MSG")
					== 0))
	    )
	    ) {
		msg = msg_parse(options);
		if (msg == NULL) {
			options->destructor(options);
			return 1;
		} else if (msg->content == NULL) {
			if (options->emailonly) {
				fprintf(stderr, "%s: %s\n",
					options->program_name,
					_("message too large"));
			} else {
				msg_dump(msg);
			}
			msg_free(msg);
			while (!feof(stdin)) {
				got =
				    fread(buffer, 1, sizeof(buffer),
					  stdin);
				if (got > sizeof(buffer))
					got = sizeof(buffer);
				if (options->emailonly)
					continue;
				totsent = 0;
				if (got > 0) {
					sent = fwrite(buffer + totsent, 1,
						      got, stdout);
					if (sent <= 0)
						break;
					totsent += sent;
					got -= sent;
				} else {
					break;
				}
			}
			options->destructor(options);
			return (options->emailonly ? 1 : 0);
		}
	}

	/*
	 * If we're just tokenising, just do that with the message we parsed
	 * above, then exit.
	 */
	if (options->tokens) {
		msg = msg_parse(options);
		if (msg == NULL) {
			options->destructor(options);
			return 1;
		}
		retcode = spam_dumptokens(options, msg);
		msg_free(msg);
		return retcode;
	}

	atexit(log_free);

	log_add(2, _("version %s initialising"), VERSION);

	/*
	 * From this point on, we're going to need databases; if none was
	 * specified on the command line, we open both /var/lib/qsfdb and
	 * ~/.qsfdb.
	 */

	dbr1 = NULL;			    /* first database to read from */
	dbr2 = NULL;			    /* second database to read from */
	dbr3 = NULL;			    /* third database to read from */
	dbw = NULL;			    /* database to write to (=dbr1 or dbr2 or NULL) */

	if (options->dump || options->restore || options->train
	    || options->mark_spam || options->mark_nonspam
	    || options->prune || options->merge)
		needwrite = 1;

	if (options->benchmark) {

#ifdef P_tmpdir
		sprintf(filename, "%.*s", (int) (sizeof(filename) - 1),
			P_tmpdir "/qsfXXXXXX");
#else
		sprintf(filename, "%.*s", (int) (sizeof(filename) - 1),
			"/tmp/qsfXXXXXX");
#endif

#ifdef HAVE_MKSTEMP
		fd = mkstemp(filename);
#else
		fd = -1;
		if (tmpnam(filename) != NULL) {	/* RATS: ignore (OK) */
			fd = open(filename, /* RATS: ignore (OK) */
				  O_RDWR | O_CREAT | O_EXCL,
				  S_IRUSR | S_IWUSR);
		}
#endif
		if (fd < 0) {
			fprintf(stderr, "%s: %s: %s: %s\n",
				options->program_name,
				filename,
				_("failed to create temporary file"),
				strerror(errno));
			retcode = 1;
		} else {
			chmod(filename,	    /* RATS: ignore (not important) */
			      S_IRUSR | S_IWUSR);
			dbr1 = qdb_open(filename, QDB_READWRITE);
			dbw = dbr1;
			if (dbr1 == NULL) {
				log_add(1, "%s: %s: %s", filename,
					_("failed to open database"),
					qdb_error());

				fprintf(stderr, "%s: %s: %s: %s\n",
					options->program_name,
					filename,
					_("failed to open database"),
					qdb_error());
				retcode = 1;
			} else {
				log_add(2, "%s: %s",
					_("using database (rw)"),
					filename);
			}
			close(fd);
			remove(filename);   /* RATS: ignore (no race) */
		}

	} else if (options->database == NULL) {

		if (options->globaldb) {
			sprintf(buffer, "%.*s", (int) (sizeof(buffer) - 1),
				options->globaldb);
		} else {
			sprintf(buffer,
				"%.*s",
				(int) (sizeof(buffer) - 1),
				"/var/lib/" PACKAGE "db");
		}

		if (needwrite) {
			if (options->restore) {
				dbr1 = qdb_open(buffer, QDB_NEW);
			} else {
				dbr1 = qdb_open(buffer, QDB_READWRITE);
			}
		}

		if (dbr1 == NULL) {
			dbr1 = qdb_open(buffer, QDB_READONLY);
			if (dbr1)
				log_add(2, "%s: %s",
					_("using database (ro)"), buffer);
		} else {
			dbw = dbr1;
			log_add(2, "%s: %s", _("using database (rw)"),
				buffer);
		}

		home = getenv("HOME");	    /* RATS: ignore (sanitised) */
		if (home == NULL)
			home = "/";

		if (strlen(home) /* RATS: ignore */ >(sizeof(buffer) - 64))
			home = "/";

#if HAVE_SNPRINTF
		snprintf(buffer, sizeof(buffer) - 1,	/* RATS: ignore (OK) */
			 "%s/.%sdb", home, PACKAGE);
#else
		sprintf(buffer,		    /* RATS: ignore (checked above) */
			"%s/.%.*sdb", home,
			(int) (sizeof(buffer) - 8 -
			       strlen(home) /* RATS: ignore */ ),
			PACKAGE);
#endif

		if (needwrite) {
			if (options->restore) {
				dbr2 = qdb_open(buffer, QDB_NEW);
			} else {
				dbr2 = qdb_open(buffer, QDB_READWRITE);
			}
		}

		if (dbr2 == NULL) {
			dbr2 = qdb_open(buffer, QDB_READONLY);
			if (dbr2)
				log_add(2, "%s: %s",
					_("using database (ro)"), buffer);
		} else if (dbw == NULL) {
			dbw = dbr2;
			log_add(2, "%s: %s", _("using database (rw)"),
				buffer);
		}

		if (options->globaldb2) {
			sprintf(buffer, "%.*s", (int) (sizeof(buffer) - 1),
				options->globaldb2);
		} else {
			sprintf(buffer, "%.*s",
				(int) (sizeof(buffer) - 1),
				"/var/lib/" PACKAGE "db2");
		}

		if (dbr3 == NULL) {
			dbr3 = qdb_open(buffer, QDB_READONLY);
			if (dbr3)
				log_add(2, "%s: %s",
					_("using database (ro)"), buffer);
		}

	} else {

		dbr1 = qdb_open(options->database, QDB_READWRITE);

		if (dbr1 == NULL) {
			dbr1 = qdb_open(options->database, QDB_READONLY);
			if (dbr1)
				log_add(2, "%s: %s",
					_("using database (ro)"),
					options->database);
		} else {
			dbw = dbr1;
			log_add(2, "%s: %s", _("using database (rw)"),
				options->database);
		}

		if (dbr1 == NULL) {
			log_add(1, "%s: %s: %s", options->database,
				_("failed to open database"), qdb_error());
			fprintf(stderr, "%s: %s: %s: %s\n",
				options->program_name, options->database,
				_("failed to open database"), qdb_error());
			retcode = 1;
		}

		if (options->globaldb) {
			dbr2 = qdb_open(options->globaldb, QDB_READONLY);
			if (dbr2 == NULL) {
				log_add(1, "%s: %s: %s", options->globaldb,
					_("failed to open database"),
					qdb_error());
				fprintf(stderr, "%s: %s: %s: %s\n",
					options->program_name,
					options->globaldb,
					_("failed to open database"),
					qdb_error());
				retcode = 1;
			}
		}

		if (options->globaldb2) {
			dbr3 = qdb_open(options->globaldb2, QDB_READONLY);
			if (dbr3 == NULL) {
				log_add(1, "%s: %s: %s",
					options->globaldb2,
					_("failed to open database"),
					qdb_error());
				fprintf(stderr, "%s: %s: %s: %s\n",
					options->program_name,
					options->globaldb2,
					_("failed to open database"),
					qdb_error());
				retcode = 1;
			}
		}
	}

	options->dbr1 = dbr1;
	options->dbr2 = dbr2;
	options->dbr3 = dbr3;
	options->dbw = dbw;

	if (retcode) {
		msg_dump(msg);
		options->destructor(options);
		qdb_close(dbr1);
		qdb_close(dbr2);
		qdb_close(dbr3);
		return retcode;
	} else if (options->dump) {
		retcode = spam_db_dump(options);
		options->destructor(options);
		qdb_close(dbr1);
		qdb_close(dbr2);
		qdb_close(dbr3);
		return retcode;
	} else if (options->restore) {
		retcode = spam_db_restore(options);
		options->destructor(options);
		qdb_close(dbr1);
		qdb_close(dbr2);
		qdb_close(dbr3);
		return retcode;
	} else if (options->prune) {
		retcode = spam_db_prune(options);
		if (retcode == 0) {
			printf("%s", _("Optimising database..."));
			qdb_optimise(dbw);
			printf(" %s\n", _("done"));
		}
		options->destructor(options);
		qdb_close(dbr1);
		qdb_close(dbr2);
		qdb_close(dbr3);
		return retcode;
	} else if (options->train) {
		retcode = spam_train(options);
		options->destructor(options);
		qdb_close(dbr1);
		qdb_close(dbr2);
		qdb_close(dbr3);
		return retcode;
	} else if (options->benchmark) {
		retcode = spam_benchmark(options);
		options->destructor(options);
		qdb_close(dbr1);
		remove(filename);	    /* RATS: ignore (no race) */
		return retcode;
	} else if (options->merge) {
		retcode = spam_db_merge(options);
		options->destructor(options);
		/*
		 * Note we are closing options->dbr1 etc instead of our dbr1
		 * etc because spam_db_merge() shuffles them around.
		 */
		qdb_close(options->dbr1);
		qdb_close(options->dbr2);
		qdb_close(options->dbr3);
		return retcode;
	} else if (options->emailonly) {
		if (strcmp(options->emailonly, "MSG") == 0) {
			if ((msg) && (msg->sender))
				options->emailonly = msg->sender;
		}
		retcode = spam_allowlist_manage(options);
		options->destructor(options);
		qdb_close(dbr1);
		qdb_close(dbr2);
		qdb_close(dbr3);
		if (msg)
			msg_free(msg);
		return retcode;
	}

	if (options->mark_spam) {
		retcode = spam_update(options, msg, SPAM);
		msg_free(msg);
	} else if (options->mark_nonspam) {
		retcode = spam_update(options, msg, NONSPAM);
		msg_free(msg);
	} else {

		score = spam_check(options, msg);

		log_add(2, _("raw score: %g"), score);

		if (!options->no_header)
			msg_spamheader(msg, score);
		if (options->add_rating) {
			msg_spamratingheader(msg, score,
					     options->threshold);
			if (options->no_filter) {
				double spamscore, scaledscore;

				spamscore = score;
				if (spamscore < 0)
					spamscore += 0.01;
				spamscore += options->threshold;
				scaledscore = spamscore * 100.0;
				printf("%d\n", (int) scaledscore);
			}
		}
		if (options->add_stars)
			msg_spamlevelheader(msg, score, options->threshold);
		if (score > 0) {
			if (options->modify_subject)
				msg_spamsubject(msg,
						options->subject_marker);
			if (!options->no_filter) {
				msg_dump(msg);
				retcode = 0;
			} else {
				retcode = 1;
			}
		} else {
			if (!options->no_filter)
				msg_dump(msg);
			retcode = 0;
		}
		msg_free(msg);
	}

	options->destructor(options);

	qdb_close(dbr1);
	qdb_close(dbr2);
	qdb_close(dbr3);

	return retcode;
}

/* EOF */
