/*
 * Parse command-line options.
 *
 * Copyright 2004 Andrew Wood, distributed under the Artistic License.
 */

#include "config.h"
#include "options.h"
#include "log.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include <unistd.h>

#ifndef HAVE_GETOPT

int minigetopt(int, char **, char *);
extern char *minioptarg;
extern int minioptind, miniopterr, minioptopt;

#define getopt minigetopt
#define optarg minioptarg
#define optind minioptind
#define opterr miniopterr
#define optopt minioptopt

#endif				/* !HAVE_GETOPT */


void display_help(void);
void display_license(void);
void display_version(void);


/*
 * Free an opts_t object.
 */
void parse_options__destroy(opts_t object)
{
	if (!object)
		return;
	if (object->argv)
		free(object->argv);
	free(object);
}


/*
 * Parse the given command-line arguments into an opts_t object, handling
 * "help", "license" and "version" options internally.
 *
 * Returns an opts_t, or 0 on error.
 *
 * Note that the contents of *argv[] (i.e. the command line parameters)
 * aren't copied anywhere, just the pointers are copied, so make sure the
 * command line data isn't overwritten or argv[1] free()d or whatever.
 */
opts_t parse_options(int argc, char **argv)
{
#ifdef HAVE_GETOPT_LONG
	struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"license", 0, 0, 'l'},
		{"version", 0, 0, 'V'},
		{"subject", 0, 0, 's'},
		{"subject-marker", 1, 0, 'S'},
		{"no-header", 0, 0, 'n'},
		{"rate", 0, 0, 'r'},
		{"rating", 0, 0, 'r'},
		{"add-rating", 0, 0, 'r'},
		{"asterisk", 0, 0, 'A'},
		{"asterisks", 0, 0, 'A'},
		{"stars", 0, 0, 'A'},
		{"add-stars", 0, 0, 'A'},
		{"add-asterisk", 0, 0, 'A'},
		{"add-asterisks", 0, 0, 'A'},
		{"test", 0, 0, 't'},
		{"allowlist", 0, 0, 'a'},
		{"allow-list", 0, 0, 'a'},
		{"level", 1, 0, 'L'},
		{"threshold", 1, 0, 'L'},
		{"email", 1, 0, 'e'},
		{"email-only", 1, 0, 'e'},
		{"mark-spam", 0, 0, 'm'},
		{"mark-nonspam", 0, 0, 'M'},
		{"mark-non-spam", 0, 0, 'M'},
		{"database", 1, 0, 'd'},
		{"global", 1, 0, 'g'},
		{"weight", 1, 0, 'w'},
		{"train", 0, 0, 'T'},
		{"noprune", 0, 0, 'N'},
		{"no-prune", 0, 0, 'N'},
		{"no-autoprune", 0, 0, 'N'},
		{"noautoprune", 0, 0, 'N'},
		{"prune", 0, 0, 'p'},
		{"trim", 0, 0, 'p'},
		{"dump", 0, 0, 'D'},
		{"restore", 0, 0, 'R'},
		{"tokens", 0, 0, 'O'},
		{"benchmark", 0, 0, 'B'},
		{"merge", 1, 0, 'E'},
		{"verbose", 0, 0, 'v'},
		{0, 0, 0, 0}
	};
	int option_index = 0;
#endif
	char *short_options = "hlVsS:nrAtoaL:e:mMd:g:w:NTpDROBE:v";
	int c;
	opts_t options;

	options = calloc(1, sizeof(*options));
	if (!options) {
		fprintf(stderr,		    /* RATS: ignore (OK) */
			_("%s: option structure allocation failed (%s)"),
			argv[0], strerror(errno));
		fprintf(stderr, "\n");
		return 0;
	}

	options->program_name = argv[0];

	options->destructor = parse_options__destroy;

	options->argc = 0;
	options->argv = calloc(argc + 1, sizeof(char *));
	if (!options->argv) {
		fprintf(stderr,		    /* RATS: ignore (OK) */
			_
			("%s: option structure argv allocation failed (%s)"),
			argv[0], strerror(errno));
		fprintf(stderr, "\n");
		options->destructor(options);
		return 0;
	}

	options->threshold = 0.9;

	do {
#ifdef HAVE_GETOPT_LONG
		c = getopt_long(argc, argv, /* RATS: ignore */
				short_options, long_options,
				&option_index);
#else
		c = getopt(argc, argv, short_options);	/* RATS: ignore */
#endif

		if (c < 0)
			continue;

		switch (c) {
		case 'h':
			display_help();
			options->do_nothing = 1;
			return options;
			break;
		case 'l':
			display_license();
			options->do_nothing = 1;
			return options;
			break;
		case 'V':
			display_version();
			options->do_nothing = 1;
			return options;
			break;
		case 's':
			options->modify_subject = 1;
			break;
		case 'S':
			options->modify_subject = 1;
			options->subject_marker = optarg;
			break;
		case 'n':
			options->no_header = 1;
			break;
		case 'r':
			options->add_rating = 1;
			break;
		case 'A':
			options->add_stars = 1;
			break;
		case 't':
			options->no_filter = 1;
			break;
		case 'a':
			options->allowlist = 1;
			break;
		case 'L':
			options->threshold = (double) (atoi(optarg)) / 100;
			break;
		case 'e':
			options->emailonly = optarg;
			break;
		case 'm':
			options->mark_spam = 1;
			break;
		case 'M':
			options->mark_nonspam = 1;
			break;
		case 'd':
			options->database = optarg;
			break;
		case 'g':
			if (options->globaldb) {
				options->globaldb2 = optarg;
			} else {
				options->globaldb = optarg;
			}
			break;
		case 'w':
			options->weight = atoi(optarg);
			break;
		case 'T':
			options->train = 1;
			break;
		case 'N':
			options->noautoprune = 1;
			break;
		case 'p':
			options->prune = 1;
			break;
		case 'D':
			options->dump = 1;
			break;
		case 'R':
			options->restore = 1;
			break;
		case 'O':
			options->tokens = 1;
			break;
		case 'B':
			options->benchmark = 1;
			break;
		case 'E':
			options->merge = 1;
			options->mergefrom = optarg;
			break;
		case 'v':
			options->loglevel++;
			break;
		default:
#ifdef HAVE_GETOPT_LONG
			fprintf(stderr,	    /* RATS: ignore (OK) */
				_("Try `%s --help' for more information."),
				argv[0]);
#else
			fprintf(stderr,	    /* RATS: ignore (OK) */
				_("Try `%s -h' for more information."),
				argv[0]);
#endif
			fprintf(stderr, "\n");
			options->destructor(options);
			return 0;
			break;
		}

	} while (c != -1);

	log_level(options->loglevel);

	if (options->weight < 1) {
		options->weight = 1;
	} else if (options->weight > 8) {
		options->weight = 8;
	}

	if (options->threshold < 0.01) {
		options->threshold = 0.01;
	} else if (options->threshold > 1.00) {
		options->threshold = 1.00;
	}

	while (optind < argc) {
		options->argv[options->argc++] = argv[optind++];
	}

	if (options->train && options->argc != 2) {
		fprintf(stderr, "%s: %s\n", argv[0],
			_
			("train requires two arguments, spam and non-spam"));
		options->destructor(options);
		return 0;
	} else if (options->benchmark && options->argc != 2) {
		fprintf(stderr, "%s: %s\n", argv[0],
			_
			("benchmark requires two arguments, spam and non-spam"));
		options->destructor(options);
		return 0;
	} else if (options->dump && options->argc > 1) {
		fprintf(stderr, "%s: %s\n", argv[0],
			_
			("dump only requires one argument (file to dump to)"));
		options->destructor(options);
		return 0;
	} else if (options->restore && options->argc > 1) {
		fprintf(stderr, "%s: %s\n", argv[0],
			_
			("restore only requires one argument (file to restore from)"));
		options->destructor(options);
		return 0;
	} else if ((!options->dump) && (!options->restore)
		   && (!options->train)
		   && (options->argc > 0)) {
		fprintf(stderr, "%s: %s\n", argv[0],
			_
			("spurious extra arguments given on command line"));
		options->destructor(options);
		return 0;
	}

	return options;
}

/* EOF */
